/*
 * This file is part of MyAdFilter <https://www.myadfilter.com/>,
 * Copyright (C) 2015-2016 DYP Solutions
 *
 * MyAdFilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * MyAdFilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Adblock Plus.  If not, see <http://www.gnu.org/licenses/>.
 */

let {LimitedSiteFilter, BlacklistDomainFilter} = require("filterClasses");


function BlacklistFilterList()
{
    this.clear();
}
BlacklistFilterList.prototype = {
  /**
   * list of domains
   * domain -> true
   */
  domains: null,

  /**
   * Removes all domains
   */
  clear: function()
  {
    this.domains = Object.create(null);
  },

  add: function(filter)
  {
    if (!(filter instanceof BlacklistDomainFilter))
      return;

    let domain = filter.domain;

    if (!(domain in this.domains))
      this.domains[ domain ] = true;
  },

  remove: function(filter)
  {
    if (!(filter instanceof BlacklistDomainFilter))
      return;

    if (filter.domain in this.domains) {
      delete this.domains[ filter.domain ];
    }
  },

  getDomain: function(url)
  {
    let scheme = url.indexOf('://');
    if (scheme != -1)
      url = url.substring(scheme + 3);
    let slash = url.indexOf('/');
    if (slash != -1)
      url = url.substring(0, slash)

    return url;
  },

  hasDomain: function(domain)
  {
    if (!domain)
      return false;

    domain = this.getDomain(domain);
    if (domain in this.domains)
      return true;
    return false;
  },
}


function LimitedSiteFilterList()
{
    this.clear();
}
LimitedSiteFilterList.prototype = {
  domainsRules: null,

  /**
   * Remove all rules
   */
  clear: function()
  {
    this.domainsRules = Object.create(null);
  },

  add: function(filter)
  {
    if (!(filter instanceof LimitedSiteFilter))
      return;

    if (!(filter.domain in this.domainsRules)) {
      this.domainsRules[ filter.domain ] = {
        simpleSelectors:  Object.create(null),
        complexSelectors: Object.create(null),
        domainCssRules:   Object.create(null),
        nbFilters:        0
      };
    }

    if (filter.hasSelector()) {
      // no parent selector and no css style
      if (filter.isSimpleFilter()) {
        this.domainsRules[ filter.domain ].simpleSelectors[ filter.selector ] = false;
      } else {
        this.domainsRules[ filter.domain ].complexSelectors[ filter.selector ] = filter.parentSelector;
      }
      this.domainsRules[ filter.domain ].nbFilters++;
    } else {
      // domain related css rule
      if (filter.hasCssRule()) {
        this.domainsRules[ filter.domain ].domainCssRules[ filter.cssSelector ] = filter.cssRule;
        this.domainsRules[ filter.domain ].nbFilters++;
      }
    }
    
  },

  remove: function(filter)
  {
    if (!(filter instanceof LimitedSiteFilter))
      return;

    if (!filter.domain in this.domainsRules)
      return;

    if (filter.hasSelector()) {
      if (filter.isSimpleFilter()) {
        delete this.domainsRules[ filter.domain ].simpleSelectors[ filter.selector ];
      } else {
        delete this.domainsRules[ filter.domain ].complexSelectors[ filter.selector ];
      }
      this.domainsRules[ filter.domain ].nbFilters--;
    } else {
      if (filter.hasCssRule()) {
        delete this.domainsRules[ filter.domain ].domainCssRules[ filter.cssSelector ];
        this.domainsRules[ filter.domain ].nbFilters--;
      }
    }
   
    if (this.domainsRules[ filter.domain ].nbFilters == 0)
      delete this.domainsRules[ filter.domain ];
  },

  getDomainsRules: function(domain)
  {
    return this.domainsRules[ this.getDomain(domain) ] || {};
  },

  getDomain: function(url)
  {
    let scheme = url.indexOf('://');
    if (scheme != -1)
      url = url.substring(scheme + 3);
    let slash = url.indexOf('/');
    if (slash != -1)
      url = url.substring(0, slash)

    return url;
  },

  hasDomain: function(domain)
  {
    if (!domain)
      return false;

    domain = this.getDomain(domain);
    if (domain in this.domainsRules)
      return true;
    return false;
  }
}


/**
 * shared limited site list
 */
let globalLimitedSiteList = exports.globalSiteList = new LimitedSiteFilterList();
let globalBlacklistList = exports.globalBlacklistList = new BlacklistFilterList();